//********************************************************************
//  Unit.java
//    Represents one unit of figures.
//    Ranks = depth, files = width (frontage).
//  Copyright (c) 2009-2011 Daniel R. Collins. All rights reserved.
//********************************************************************

public class Unit {
	enum Type {infantry, archers, cavalry};
	enum Alignment {Lawful, Neutral, Chaotic};

	private String name;
	private Type type;
	private int cost;
	private int MV, AH, HD, ROF, RA;
	private int figures, damage;
	private int initFiles, figsLostInTurn;
	private double figWidth;
	private boolean routed;
	private boolean pikes;
	private Alignment alignment;
	private Hero hero;

	//-----------------------------------------------------------------
	//  Constructor basics
	//-----------------------------------------------------------------
	public Unit (String name, Type type, 
			int cost, int MV, int AH, int HD) {

		// Parameter stuff
		this.name = name;
		this.type = type;
		this.cost = cost;
		this.MV = MV;
		this.AH = AH;
		this.HD = HD;

		// Initial null stuff
		this.RA = 0;
		this.ROF = 0;
		this.initFiles = 0;
		this.figures = 0;
		this.damage = 0;
		this.figsLostInTurn = 0;
		this.routed = false;
		this.alignment = Alignment.Neutral;
		this.hero = null;

		// Check for pikes
		pikes	= name.contains("Pike");

		// Default figure sizes
		switch(type) {
			case cavalry: figWidth = 1.0; break;
			default: figWidth = 0.75; break;
		}
	}

	//-----------------------------------------------------------------
	//  Constructor for archers
	//-----------------------------------------------------------------
	public Unit(String name, Type type, 
			int cost, int MV, int AH, int HD, int ROF, int RA) {

		this(name, type, cost, MV, AH, HD);
		this.ROF = ROF;
		this.RA = RA;
	}

	//-----------------------------------------------------------------
	//  Constructor for fantasy types
	//-----------------------------------------------------------------
	public Unit(String name, Type type, Alignment alignment, 
			double width, int cost, int MV, int AH, int HD, int ROF, int RA) {

		this(name, type, cost, MV, AH, HD, ROF, RA);
		this.alignment = alignment;
		this.figWidth = width;
	}

	//-----------------------------------------------------------------
	//  Constructor for units with heroes (melee)
	//-----------------------------------------------------------------
	public Unit(String name, Type type, 
			int cost, int MV, int AH, int HD, Hero hero) {

		this(name, type, cost, MV, AH, HD);
		this.hero = hero;
	}

	//-----------------------------------------------------------------
	//  Constructor for units with heroes (archers)
	//-----------------------------------------------------------------
	public Unit(String name, Type type, 
			int cost, int MV, int AH, int HD, int ROF, int RA, Hero hero) {

		this(name, type, cost, MV, AH, HD, ROF, RA);
		this.hero = hero;
	}

	//-----------------------------------------------------------------
	//  Accessors
	//-----------------------------------------------------------------
	public String getName() { return name; };
	public Type getType() { return type; };
	public Alignment getAlignment() { return alignment; };
	public Hero getHero() { return hero; };
	public int getCost() { return cost; };
	public int getMV() { return MV; };
	public int getAH() { return AH; };
	public int getHD() { return HD; };
	public int getROF() { return ROF; };
	public int getRA() { return RA; };
	public int getFigures() { return figures; };
	public int getFigsLostInTurn() { return figsLostInTurn; };
	public double getFigWidth() { return figWidth; };
	public boolean hasMissiles() { return ROF > 0; };
	public boolean hasPikes() { return pikes; };
	public boolean hasHero() { return hero != null; };
	public boolean isBeaten() { return figures == 0 || routed; };

	//-----------------------------------------------------------------
	//  Initialize figures
	//-----------------------------------------------------------------
	public void setFigures (int figures) {
		this.figures = figures;
		this.routed = false;
		this.damage = 0;
	}
	
	//-----------------------------------------------------------------
	//  Initialize frontage by desired rows
	//-----------------------------------------------------------------
	public void setFilesByRanks (int ranks) {
		if (ranks > figures) ranks = figures;
		initFiles = (ranks > 0 ? figures/ranks : figures);
	}

	//-----------------------------------------------------------------
	//  Compute files (width/frontage)
	//-----------------------------------------------------------------
	public int getFiles() {
		return Math.min(initFiles, figures);
	}

	//-----------------------------------------------------------------
	//  Compute how many effective ranks we have.
	//-----------------------------------------------------------------
	public int getRanks() {
		int files = getFiles();
		if (files < 1) return 0;
		int ranks = figures / files;
		int backrow = figures % files;
		if (backrow*2 >= files && files > 1) ranks++;
		return ranks;	
	}

	//-----------------------------------------------------------------
	//  Compute how long the figure is (assume cavalry 2 inches)
	//-----------------------------------------------------------------
	public double getFigLength() {
		return (type == Type.cavalry ? 2.0 : figWidth);	
	}

	//-----------------------------------------------------------------
	//  Remove a certain number of figures.
	//-----------------------------------------------------------------
	private int removeFigures (int lost) {
		if (lost >= figures) {
			lost = figures;
			damage = 0;		
		}
		figures -= lost;
		figsLostInTurn += lost;
		return lost;
	}

	//-----------------------------------------------------------------
	//  Take damage; returns figures killed.
	//-----------------------------------------------------------------
	public int takeDamage (int points) {
		assert(points >= 0);		
		damage += points;
		int figsKilled = damage/HD;
		damage = damage % HD;
		return removeFigures(figsKilled);
	}

	//-----------------------------------------------------------------
	//  Clear figures lost in turn.
	//-----------------------------------------------------------------
	public void clearFigsLostInTurn () {
		figsLostInTurn = 0;	
	}

	//-----------------------------------------------------------------
	//  Regenerate (for trolls)
	//    Each figure in contact gets back 1 hit
	//-----------------------------------------------------------------
	public void regenerate () {
		damage -= getFiles(); 
		if (damage < 0) damage = 0;
	}

	//-----------------------------------------------------------------
	//  Is this unit a wizard(s)?
	//-----------------------------------------------------------------
	public boolean isWizard () {
		return (this instanceof Hero && ((Hero)this).isWizard());
	}

	//-----------------------------------------------------------------
	//  Is this unit wizard(s), or have an attached wizard?
	//-----------------------------------------------------------------
	public boolean hasAnyWizard () {
		return (isWizard() || hasHero() && getHero().isWizard());
	}
	
	//-----------------------------------------------------------------
	//  Set routed status
	//-----------------------------------------------------------------
	public void setRouted (boolean routed) {
		this.routed = routed;
	}

	//-----------------------------------------------------------------
	//  Does name start with this string?
	//-----------------------------------------------------------------
	public boolean nameStarts (String s) {
		return name.startsWith(s);
	}

	//-----------------------------------------------------------------
	//  Does name contain this string?
	//-----------------------------------------------------------------
	public boolean nameContains (String s) {
		return name.contains(s);
	}

	//-----------------------------------------------------------------
	//  Pluralize something with an "s"
	//-----------------------------------------------------------------
	public String plural (int n) {
		return (n == 1 ? "" : "s");
	}

	//-----------------------------------------------------------------
	//  Returns a string representation of this unit.
	//-----------------------------------------------------------------
	public String toString() {
		return name + " (" + figures + " fig" + plural(figures) + ", "
			+ getRanks() + " rank" + plural(getRanks()) + ")";
	}
	
	//-----------------------------------------------------------------
	//  Returns an abbreviation for this unit.
	//-----------------------------------------------------------------
	public String getAbbreviation() {
		String s = "" + name.charAt(0);
		for (int i = 1; i < name.length()-1; i++) {
			if (name.charAt(i) == ' ' && s.length() < 3)
				s = s + name.charAt(i+1);
		}
		return s;
	}
}
